// drweb-clamd API usage example.
// (C) 2003-2016 Doctor Web. All rights reserved.

#include <algorithm>
#include <cstring>
#include <iostream>

#include <arpa/inet.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/un.h>
#include <unistd.h>

#include "drweb_clamd_api_common.hpp"

static std::string make_cmd_list(const std::vector<std::string>& cmds)
{
    std::ostringstream os;
    auto it = cmds.cbegin();
    if(it != cmds.cend()) os << *it++;
    while(it != cmds.cend()) os << " | " << *it++;
    return os.str();
}

parsed_args_t parse_args(int argc, const char* argv[], const std::vector<std::string>& supported_cmds)
{
    auto throw_usage = [&]() { THROW("Usage: " << argv[0] << " <socket path> <files...> [-c " << make_cmd_list(supported_cmds) << ']'); };

    if(argc < 3) throw_usage();

    parsed_args_t args;
    args.socket_path = argv[1];

    for(int i = 2; i < argc; ++i) {
        if(std::strcmp(argv[i], "-c") == 0) {
            if(++i == argc) throw_usage();
            if(std::find(supported_cmds.begin(), supported_cmds.end(), argv[i]) == supported_cmds.end())
                throw_usage();
            args.command = argv[i];
        }
        else args.files.push_back(argv[i]);
    }

    if(args.files.empty()) throw_usage();

    if(args.command.empty() && !supported_cmds.empty())
        args.command = supported_cmds.front();

    return args;
}

int connect_to_clamd(const char* socket_path)
{
    scoped_fd_t sock = socket(AF_LOCAL, SOCK_STREAM, 0);
    if(sock == -1) THROW("ERROR: socket: " << strerror(errno));

    sockaddr_un addr;
    std::memset(&addr, 0, sizeof(addr));
    addr.sun_family = AF_LOCAL;
    constexpr auto max_len = sizeof(addr.sun_path) / sizeof(*addr.sun_path) - 1;
    if(strlen(socket_path) > max_len)
        THROW("ERROR: socket path is too long: " << socket_path);
    std::strcpy(addr.sun_path, socket_path);

    if(connect(sock, reinterpret_cast<sockaddr*>(&addr), sizeof(addr)) == -1)
        THROW("ERROR: connect(" << socket_path << "): " << strerror(errno));

    return sock.release();
}

void write_exact(int fd, const void* buffer, size_t size)
{
    ssize_t total = 0;
    do {
        ssize_t part = write(fd, static_cast<const uint8_t*>(buffer) + total, size - total);
        if(part > 0)
            total += part;
        else if(part == 0 && size > 0)
            THROW("ERROR: write: unexpected error");
        else if(part == -1) {
            if(errno == EINTR) continue;
            THROW("ERROR: write: " << strerror(errno));
        }
    } while(total > 0 && static_cast<size_t>(total) < size);
}

void read_response(int fd)
{
    char buf[4096];
    while(true) {
        ssize_t size = read(fd, buf, sizeof(buf) / sizeof(*buf));
        if(size == -1) {
            if(errno == EINTR) continue;
            THROW("ERROR: read: " << strerror(errno));
        }
        else if(size == 0) break;
        std::cout.write(buf, size);
    }
}

std::string read_until(int sock, char term)
{
    std::string result;
    char buf[4096];
    while(true) {
        ssize_t size = recv(sock, buf, sizeof(buf) / sizeof(*buf), MSG_PEEK);
        if(size == -1) {
            if(errno == EINTR) continue;
            THROW("ERROR: recv: " << strerror(errno));
        }
        else if(size == 0) break;

        char* pos = std::find(buf, buf + size, term);
        ssize_t consume = pos - buf;
        bool found = consume < size;
        if(found) ++consume;
        recv(sock, buf, consume, 0);
        result.append(buf, consume);
        if(found) break;
    }
    return result;
}

void send_by_chunks(int clamd, int fd)
{
    char buf[4096];
    while(true) {
        ssize_t size = read(fd, buf + 4, sizeof(buf) / sizeof(*buf) - 4);
        if(size == -1) {
            if(errno == EINTR) continue;
            THROW("ERROR: read: " << strerror(errno));
        } else if(size == 0) {
            write_exact(clamd, "\x00\x00\x00\x00", 4);
            break;
        }
        uint32_t* header = reinterpret_cast<uint32_t*>(buf);
        *header = htonl(size);
        write_exact(clamd, buf, size + 4);
    }
}

void send_fd(int clamd, int fd)
{
    iovec iov;
    char garbage = '\x00';
    iov.iov_base = &garbage;
    iov.iov_len = 1;

    char cmsgbuf[CMSG_SPACE(sizeof(fd))];

    msghdr mh;
    std::memset(&mh, 0, sizeof(mh));
    mh.msg_iov = &iov;
    mh.msg_iovlen = 1;
    mh.msg_control = cmsgbuf;
    mh.msg_controllen = sizeof(cmsgbuf);

    cmsghdr* cmsg = CMSG_FIRSTHDR(&mh);
    cmsg->cmsg_level = SOL_SOCKET;
    cmsg->cmsg_type = SCM_RIGHTS;
    cmsg->cmsg_len = CMSG_LEN(sizeof(int));
    int* payload = reinterpret_cast<int*>(CMSG_DATA(cmsg));
    *payload = fd;

    mh.msg_controllen = cmsg->cmsg_len;

    int result = sendmsg(clamd, &mh, 0);
    if(result == -1)
        THROW("ERROR: sendmsg: " << strerror(errno));
    else if(result == 0)
        THROW("ERROR: sendmsg: unexpected error");
}



void scan(int clamd, const std::string& file, const std::string& command)
{
    std::string cmd;
    {
        std::ostringstream os;
        os << 'n' << command << ' ' << file << '\n';
        cmd = os.str();
    }
    write_exact(clamd, cmd.c_str(), cmd.size());
}

void instream_scan(int clamd, const std::string& path)
{
    std::string cmd = "nINSTREAM\n";
    write_exact(clamd, cmd.c_str(), cmd.size());

    scoped_fd_t file = open(path.c_str(), O_RDONLY);
    if(file == -1) THROW("ERROR: open(" << path << "): " << strerror(errno));

    send_by_chunks(clamd, file);
}

void fildes_scan(int clamd, const std::string& path)
{
    std::string cmd = "nFILDES\n";
    write_exact(clamd, cmd.c_str(), cmd.size());

    scoped_fd_t file = open(path.c_str(), O_RDONLY);
    if(file == -1) THROW("ERROR: open(" << path << "): " << strerror(errno));

    send_fd(clamd, file);
}
